import com.apama.event.Event;
import com.apama.event.parser.EventType;
import com.apama.event.parser.FieldTypes;
import com.apama.services.event.EventServiceFactory;
import com.apama.services.event.IEventService;
import com.apama.services.event.IEventServiceChannel;
import com.apama.services.event.IResponseListener;
import com.apama.util.Logger;

/*
 * ESChannelAsyncRequestResponse.java
 *
 * This sample demonstrates the asyncRequestResponse() method of the
 * IEventServiceChannel class.
 * 
 * A ResponseListener will be registered to the IEventServiceChannel and the event
 * AsyncRequestEvent(0, "Hi There") will be sent to the correlator by calling the
 * IEventServiceChannel's asyncRequestResponse() method and the AsyncResponseEvent() 
 * is expected as a response.  
 * 
 * Requirement:
 *
 * The RequestResponseEvent.mon must be injected to the correlator before
 * running this example.  Upon receiving the AsyncRequestEvent() event, the
 * AsyncRequestResponseEvent.mon monitorscript will send the response
 * (AsyncResponseEvent()) back to the same channel and hence the callback of the 
 * ResponseListener will be called.
 *
 * $Copyright(c) 2005 - 2006, 2009-2010, 2012 Progress Software Corporation (PSC). All rights reserved.$
 * $Copyright (c) 2013, 2015, 2017 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 *
 */

public class ESChannelAsyncRequestResponse {
	private static final Logger logger = Logger.getLogger(ESChannelAsyncRequestResponse.class);
	
	private volatile boolean done = false;
	
	public void runSample(String host, int port) throws Exception 
	{

		try
		(
			// get a IEventService instance from the EventServiceFactory
			// IEventService.close() will automatically get called upon exiting the try-with-resources block
			IEventService eventService = EventServiceFactory.createEventService(host, port, "ESChannelAsyncRequestResponse sample");
		) {
			// Create the channel on "eventService.sample.channel"
			IEventServiceChannel ourChannel = eventService.addChannel(new String[]{"eventService.sample.channel"}, null);
			
			// Define the AsyncRequestEvent and AsyncResponseEvent event type
			// NOTE: in order to use the asyncRequestResponse() method, the Request
			//       and Response event must have a member  "messageId" 
			//		 (IEventServiceChannel.DEFAULT_MESSAGEID_FIELD_NAME) of 
			//       MonitorScript type integer, Java type Long.
			EventType asyncRequestEventType = new EventType("AsyncRequestEvent",
					FieldTypes.INTEGER.newField(IEventServiceChannel.DEFAULT_MESSAGEID_FIELD_NAME), // field required for using asyncRequestResponse() call
					FieldTypes.STRING.newField("msg")
					);
			
			EventType asyncResponseEventType = new EventType("AsyncResponseEvent",
					FieldTypes.INTEGER.newField(IEventServiceChannel.DEFAULT_MESSAGEID_FIELD_NAME),  // field required for using asyncRequestResponse() call
					FieldTypes.STRING.newField("responseMsg"));
			
			// Register the RequestEventType to the channel
			ourChannel.registerEventType(asyncRequestEventType);
			
			// Create the requestEvent
			Event requestEvent = new Event("AsyncRequestEvent(0, \"Hi There\")");
			
			// Now, make the asyncRequestReponse.  MyReponseListener's 
			// handleResponse() is the callback method when a response is available
			ourChannel.asyncRequestResponse(new MyResponseListener(),
											requestEvent,
											asyncResponseEventType);
			
			System.out.println("AsyncRequestEvent() sent...");
	
			while(!done) Thread.sleep(500);
		}
	}
	
	/** Inner class to implement our callback method that is notified when events are recieved. */
	private class MyResponseListener implements IResponseListener 
	{

		/** Callback method that is called when a matching response Event is 
		 * received for an asynchronous request-response call made on an EventServiceChannel.
		 * 
		 * 
		 * @param requestEvent the original request Event.
		 * @param responseEvent the corresponding response Event.
		 */
		public void handleResponse(Event requestEvent, Event responseEvent) {
			System.out.println("MyResponseListener.handleResponse() called - ");
			System.out.println("\t requestEvent = " + requestEvent);
			System.out.println("\t responseEvent = " + responseEvent);
			
			done = true;
		}
		
		
		/** Callback method that is called when an exception is thrown within an
		 * asynchronous request-response call made on an EventServiceChannel.
		 * 
		 * @param exception The exception that was thrown in the asynchronous call.
		 */ 
		public void handleException(Exception exception) {
			logger.error("MyResponseListener got an error: ", exception);
			// set it to true so that program exits in case of failure
			done = true;
		}
	}
	

	public static void main(String[] args) throws Exception
	{
		if ((args.length != 2) || (Integer.parseInt(args[1]) <= 0)) {
			// Bad command line given
			System.out.println("Usage: java ESChannelAsyncRequestResponse <host> <port>");
			return;
		}
		String host = args[0];
		int port = Integer.parseInt(args[1]);
		new ESChannelAsyncRequestResponse().runSample(host, port);
	}

}
