import com.apama.event.Event;
import com.apama.event.parser.*;
import com.apama.services.event.*;

/*
 * ESChannelRequestResponse.java
 *
 * This sample demonstrates the requestResponse() method of the
 * IEventServiceChannel class.
 * 
 * A ResponseListener will be registered to the IEventServiceChannel and the event
 * RequestEvent(0, "Hi There") will be sent to the correlator by calling the
 * IEventServiceChannel's requestResponse() method and the ResponseEvent() 
 * is expected as a response.  
 * 
 * Requirement:
 *
 * The RequestResponseEvent.mon must be injected to the correlator before
 * running this example.  Upon receiving the RequestEvent() event, the
 * RequestResponseEvent.mon monitorscript will send the response
 * (ResponseEvent()) back to the same channel and hence the callback of the 
 * ResponseListener will be called.
 *
 * $Copyright(c) 2005 - 2006, 2009-2010, 2012 Progress Software Corporation (PSC). All rights reserved.$
 * $Copyright (c) 2013, 2015, 2017 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 *
 */

public class ESChannelRequestResponse {

	// Define the RequestEvent and ResponseEvent event type
	// NOTE: in order to use the requestResponse() method, the Request
	//       and Response event must have a member  "messageId" 
	//		 (IEventServiceChannel.DEFAULT_MESSAGEID_FIELD_NAME) of 
	//       MonitorScript type integer, Java type Long.
	
	static final EventType EVENT_REQUEST = new EventType("RequestEvent",
			FieldTypes.INTEGER.newField(IEventServiceChannel.DEFAULT_MESSAGEID_FIELD_NAME),  // field required for using requestResponse() call
			FieldTypes.STRING.newField("msg")
			);
	
	static final EventType EVENT_RESPONSE = new EventType("ResponseEvent",
			FieldTypes.INTEGER.newField(IEventServiceChannel.DEFAULT_MESSAGEID_FIELD_NAME),  // field required for using requestResponse() call
			FieldTypes.STRING.newField("responseMsg")
			);
	

	public void runSample(String host, int port) throws Exception {

		try
		(
			// get a IEventService instance from the EventServiceFactory
			// IEventService.close() will automatically get called upon exiting the try-with-resources block
			IEventService eventService = EventServiceFactory.createEventService(host, port, "ESChannelRequestResponse sample");
		) {
			// Create the channel on "eventService.sample.channel"
			IEventServiceChannel ourChannel = eventService.addChannel(new String[]{"eventService.sample.channel"}, null);
			
			// Register the RequestEventType to the channel
			ourChannel.registerEventType(EVENT_REQUEST);
			
			// Create the requestEvent
			Event requestEvent = new Event("RequestEvent(0, \"Hi There\")");
			
			
			// Send two events in a for loop
			for (int i = 0; i < 2; i++) {
				System.out.println("Sending RequestEvent() ...");

				try
				(
					// IResponseWrapper.close() will automatically get called upon exiting the try-with-resources block
					// which in turns calls IResponseWrapper.releaseLock (which must be released after a response is received)
					IResponseWrapper responseWrapper = ourChannel.requestResponse(requestEvent, EVENT_RESPONSE);
				) {
					// print the response event
					System.out.println("Response event received: " + responseWrapper.getEvent());
					
				}
			}
		}
	}
	
	public static void main(String[] args) throws Exception 
	{
		if ((args.length != 2) || (Integer.parseInt(args[1]) <= 0)) {
			// Bad command line given
			System.out.println("Usage: java ESChannelRequestResponse <host> <port>");
			return;
		}
		String host = args[0];
		int port = Integer.parseInt(args[1]);
		new ESChannelRequestResponse().runSample(host, port);
	}

}
